import { PropertyAnnotation } from './annotation-shapes';
import { DocumentSchema, METADATA_TYPE_KEY } from './protocols';
import { BinarySet } from './binary-set';
import { SchemaType, SetType, DocumentType } from './schema-type';
import { NumberValueSet } from './number-value-set';
import { Schema } from './schema';
import "reflect-metadata";


export function attribute(parameters: Partial<SchemaType> = {}): PropertyAnnotation {
  // eslint-disable-next-line @typescript-eslint/ban-types
  return (target: Object, propertyKey: string | symbol): void => {
    if (!Object.prototype.hasOwnProperty.call(target, DocumentSchema)) {
      Object.defineProperty(
        target,
          DocumentSchema as any, // TypeScript complains about the use of symbols here, though it should be allowed
        {value: deriveBaseSchema(target)}
      );
    }
    (target as any)[DocumentSchema][propertyKey] = metadataToSchemaType(Reflect.getMetadata(METADATA_TYPE_KEY, target, propertyKey), parameters);
  }
}

function metadataToSchemaType(
  ctor: {new (): any}|undefined,
  declaration: Partial<SchemaType>
): any {
  // eslint-disable-next-line prefer-const
  let {type, ...rest} = declaration;
  if (type === undefined) {
    if (ctor) {
      if (ctor === String) {
        type = 'String';
      } else if (ctor === Number) {
        type = 'Number';
      } else if (ctor === Boolean) {
        type = 'Boolean';
      } else if (ctor === Date || ctor.prototype instanceof Date) {
        type = 'Date';
      } else if (
        ctor === BinarySet ||
        ctor.prototype instanceof BinarySet
      ) {
        type = 'Set';
        (rest as SetType).memberType = 'Binary';
      } else if (
        ctor === NumberValueSet ||
        ctor.prototype instanceof NumberValueSet
      ) {
        type = 'Set';
        (rest as SetType).memberType = 'Number';
      } else if (ctor === Set || ctor.prototype instanceof Set) {
        type = 'Set';
        if (!('memberType' in rest)) {
          throw new Error(
            'Invalid set declaration. You must specify a memberType'
          );
        }
      } else if (ctor === Map || ctor.prototype instanceof Map) {
        type = 'Map';
        if (!('memberType' in rest)) {
          throw new Error(
            'Invalid map declaration. You must specify a memberType'
          );
        }
      } else if (ctor.prototype[DocumentSchema]) {
        type = 'Document';
        (rest as DocumentType).members = ctor.prototype[DocumentSchema];
        (rest as DocumentType).valueConstructor = ctor;
      } else if (isBinaryType(ctor)) {
        type = 'Binary';
      } else if (ctor === Array || ctor.prototype instanceof Array) {
        if ('members' in declaration) {
          type = 'Tuple';
        } else if ('memberType' in declaration) {
          type = 'List';
        } else {
          type = 'Collection';
        }
      } else {
        type = 'Any';
      }
    } else {
      type = 'Any';
    }
  }

  return {
    ...rest,
    type
  } as SchemaType;
}

function deriveBaseSchema(target: any): Schema {
  if (target && typeof target === 'object') {
    const prototype = Object.getPrototypeOf(target);
    if (prototype) {
      return {
        ...deriveBaseSchema(prototype),
        ...Object.prototype.hasOwnProperty.call(prototype, DocumentSchema)
          ? prototype[DocumentSchema]
          : {}
      };
    }
  }

  return {};
}

function isBinaryType(arg: any): boolean {
  return arg === Uint8Array || arg.prototype instanceof Uint8Array ||
    arg === Uint8ClampedArray || arg.prototype instanceof Uint8ClampedArray ||
    arg === Uint16Array || arg.prototype instanceof Uint16Array ||
    arg === Uint32Array || arg.prototype instanceof Uint32Array ||
    arg === Int8Array || arg.prototype instanceof Int8Array ||
    arg === Int16Array || arg.prototype instanceof Int16Array ||
    arg === Int32Array || arg.prototype instanceof Int32Array ||
    arg === Float32Array || arg.prototype instanceof Float32Array ||
    arg === Float64Array || arg.prototype instanceof Float64Array ||
    arg === ArrayBuffer || arg.prototype instanceof ArrayBuffer ||
    arg === DataView || arg.prototype instanceof DataView;
}
