/**
 * A constructor that takes no arguments.
 */
import { BinaryValue } from './binary-set';

export interface ZeroArgumentsConstructor<T> {
    new (): T;
}

export declare const TypeTags: {
    Any: string;
    Binary: string;
    Boolean: string;
    Collection: string;
    Custom: string;
    Date: string;
    Document: string;
    Hash: string;
    List: string;
    Map: string;
    Null: string;
    Number: string;
    Set: string;
    String: string;
    Tuple: string;
};

export declare type TypeTag = keyof typeof TypeTags;

export interface BaseType<T = any> {
    /**
     * The type of node represented by this object.
     */
    type: TypeTag;
}

export interface AnyType extends BaseType {
    type: 'Any';
}
/**
 * A node used to store binary data (e.g., Buffer, ArrayBuffer, or
 * ArrayBufferView objects).
 */
export interface BinaryType extends BaseType<BinaryValue> {
    type: 'Binary';
}
/**
 * A node used to store boolean values.
 */
export interface BooleanType extends BaseType<boolean> {
    type: 'Boolean';
}
/**
 * A node used to store an untyped or mixed collection. Values provided for this
 * node will be marshalled using run-time type detection and may not be exactly
 * the same when unmarshalled.
 */
export interface CollectionType<T = {
    [key: string]: any;
}> extends BaseType<Array<T>> {
    type: 'Collection';

    /**
     * A constructor to invoke to create an object onto which the document's
     * members will be unmarshalled. If not provided, `Object.create(null)` will
     * be used.
     */
    valueConstructor?: ZeroArgumentsConstructor<T>;

    documentConstructorMapping: DocumentConstructorMapping;
}

export interface DocumentConstructorMapping<T = {
    [key: string]: any;
}> {
    attribute: string;
    mapping: {[key: string]: ZeroArgumentsConstructor<T>};
}
/**
 * A node represented by a Date object.
 */
export interface DateType extends BaseType<string | number | Date> {
    type: 'Date';
}
/**
 * A constructor that takes no arguments.
 */
export interface ZeroArgumentsConstructor<T> {
    new (): T;
}
/**
 * A node represented by its own full Schema. Marshalled as an embedded map.
 */
export interface DocumentType<T = {
    [key: string]: any;
}> extends BaseType<T> {
    type: 'Document';
    /**
     * A Schema outlining how the members of this document are to be
     * (un)marshalled.
     */
    members: any;
    /**
     * A constructor to invoke to create an object onto which the document's
     * members will be unmarshalled. If not provided, `Object.create(null)` will
     * be used.
     */
    valueConstructor?: ZeroArgumentsConstructor<T>;
}
/**
 * A node used to store a key => value mapping of mixed or untyped values.
 * Values provided for this node will be marshalled using run-time type
 * detection and may not be exactly the same when unmarshalled.
 */
export interface HashType extends BaseType<{
    [key: string]: any;
}> {
    type: 'Hash';
}
/**
 * A node used to store an array or iterable of like values, e.g.,
 * `Array<string>`.
 *
 * @see CollectionType For untyped or mixed lists
 * @see TupleType For tuples
 */
export interface ListType<E = any> extends BaseType<Array<E>> {
    type: 'List';
    /**
     * The schema node by which each member of the list should be
     * (un)marshalled.
     */
    memberType: SchemaType;
}
/**
 * A node used to store a mapping of strings to like values, e.g.,
 * `Map<string, ArrayBuffer>`.
 *
 * @see HashType For untyped of mixed hashes
 * @see DocumentType For strongly-typed documents
 */
export interface MapType<E = any> extends BaseType<Map<string, E>> {
    type: 'Map';
    memberType: SchemaType;
}
/**
 * A node used to store null values.
 */
export interface NullType extends BaseType<null> {
    type: 'Null';
}
/**
 * A node used to store a number value. Numbers should be representable as IEEE
 * 754 double precision floating point values to ensure no precision is lost
 * during (un)marshalling.
 */
export interface NumberType extends BaseType<number> {
    type: 'Number';
    versionAttribute?: boolean;
}
export interface SetType extends BaseType<Set<any>> {
    type: 'Set';
    memberType: 'String' | 'Number' | 'Binary';
}
/**
 * A node used to store a string value.
 */
export interface StringType extends BaseType<string> {
    type: 'String';
}
/**
 * A node used to store a fixed-length list of items, each of which may be of
 * a different type, e.g., `[boolean, string]`.
 */
export interface TupleType<T extends Array<any> = Array<any>> extends BaseType<T> {
    type: 'Tuple';
    members: Array<SchemaType>;
}

export declare type SchemaType = AnyType | BinaryType | BooleanType | CollectionType | DateType | DocumentType<any> | HashType | ListType | MapType | NullType | NumberType | SetType | StringType | TupleType;
